@@ -1,10 +1,14 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
+import os
+import pickle
 import time
+import warnings
 import numpy as np
 from scipy.optimize import minimize
 from graphdot.util.printer import markdown as mprint
-from ._chol_solver import CholSolver
+from graphdot.linalg.cholesky import CholSolver
+from graphdot.linalg.spectral import pinvh
 
 
 class GaussianProcessRegressor:
@@ -14,13 +18,17 @@ class GaussianProcessRegressor:
     ----------
     kernel: kernel instance
         The covariance function of the GP.
-    alpha: float > 0, default = 1e-10
+    alpha: float > 0
         Value added to the diagonal of the kernel matrix during fitting. Larger
         values correspond to increased noise level in the observations. A
         practical usage of this parameter is to prevent potential numerical
         stability issues during fitting, and ensures that the kernel matrix is
         always positive definite in the precense of duplicate entries and/or
         round-off error.
+    beta: float > 0
+        Cutoff value on the singular values for the spectral pseudoinverse
+        computation, which serves as a backup mechanism to invert the kernel
+        matrix in case if it is singular.
     optimizer: one of (str, True, None, callable)
         A string or callable that represents one of the optimizers usable in
         the scipy.optimize.minimize method.
@@ -36,10 +44,11 @@ class GaussianProcessRegressor:
         kernel to data.
     """
 
-    def __init__(self, kernel, alpha=1e-10, optimizer=None, normalize_y=False,
-                 kernel_options={}):
+    def __init__(self, kernel, alpha=1e-8, beta=1e-8, optimizer=None,
+                 normalize_y=False, kernel_options={}):
         self.kernel = kernel
         self.alpha = alpha
+        self.beta = beta
         self.optimizer = optimizer
         if optimizer is True:
             self.optimizer = 'L-BFGS-B'
@@ -83,7 +92,7 @@ class GaussianProcessRegressor:
         kernel = kernel or self.kernel
         if Y is None:
             if diag is True:
-                return kernel.diag(X, **self.kernel_options)
+                return kernel.diag(X, **self.kernel_options) + self.alpha
             else:
                 if jac is True:
                     K, J = kernel(X, eval_gradient=True, **self.kernel_options)
@@ -105,10 +114,28 @@ class GaussianProcessRegressor:
                 else:
                     return kernel(X, Y, **self.kernel_options)
 
-    def fit(self, X, y, tol=1e-4, repeat=1, theta_jitter=1.0, verbose=False):
+    def _invert(self, K):
+        try:
+            return CholSolver(K), np.prod(np.linalg.slogdet(K))
+        except np.linalg.LinAlgError:
+            try:
+                warnings.warn(
+                    'Kernel matrix singular, falling back to pseudoinverse'
+                )
+                return pinvh(
+                    K, rcond=self.beta, mode='clamp', return_nlogdet=True
+                )
+            except np.linalg.LinAlgError:
+                raise np.linalg.LinAlgError(
+                    'The kernel matrix is likely corrupted with NaNs and Infs '
+                    'because a pseudoinverse could not be computed.'
+                )
+
+    def fit(self, X, y, loss='likelihood', tol=1e-5, repeat=1,
+            theta_jitter=1.0, verbose=False):
         """Train a GPR model. If the `optimizer` argument was set while
         initializing the GPR object, the hyperparameters of the kernel will be
-        optimized using maximum likelihood estimation.
+        optimized using the specified loss function.
 
         Parameters
         ----------
@@ -116,6 +143,10 @@ class GaussianProcessRegressor:
             Input values of the training data.
         y: 1D array
             Output/target values of the training data.
+        loss: 'likelihood' or 'loocv'
+            The loss function to be minimzed during training. Could be either
+            'likelihood' (negative log-likelihood) or 'loocv' (mean-square
+            leave-one-out cross validation error).
         tol: float
             Tolerance for termination.
         repeat: int
@@ -137,8 +168,14 @@ class GaussianProcessRegressor:
 
         '''hyperparameter optimization'''
         if self.optimizer:
+
+            if loss == 'likelihood':
+                objective = self.log_marginal_likelihood
+            elif loss == 'loocv':
+                objective = self.squared_loocv_error
+
             opt = self._hyper_opt(
-                lambda theta, self=self: self.log_marginal_likelihood(
+                lambda theta, objective=objective: objective(
                     theta, eval_gradient=True, clone_kernel=False,
                     verbose=verbose
                 ),
@@ -152,95 +189,19 @@ class GaussianProcessRegressor:
                 self.kernel.theta = opt.x
             else:
                 raise RuntimeError(
-                    f'Maximum likelihood estimation did not converge, got:\n'
+                    f'Training using the {loss} loss did not converge, got:\n'
                     f'{opt}'
                 )
 
         '''build and store GPR model'''
         self.K = K = self._gramian(self.X)
-        self.Kinv = CholSolver(K)
+        self.Kinv, _ = self._invert(K)
         self.Ky = self.Kinv @ self.y
         return self
 
-    def fit_loocv(self, X, y, return_mean=False, return_std=False,
-                  tol=1e-4, repeat=1, theta_jitter=1.0, verbose=False):
-        """Train a GPR model and return the leave-one-out cross validation
-        results on the dataset. If the `optimizer` argument was set while
-        initializing the GPR object, the hyperparameters of the kernel will be
-        optimized with regard to the LOOCV RMSE.
-
-        Parameters
-        ----------
-        X: list of objects or feature vectors.
-            Input values of the training data.
-        y: 1D array
-            Output/target values of the training data.
-        return_mean: boolean
-            If True, the leave-one-out predictive mean of the model on the
-            training data are returned along with the model.
-        return_std: boolean
-            If True, the leave-one-out predictive standard deviations of the
-            model on the training data are returned along with the model.
-        tol: float
-            Tolerance for termination.
-        repeat: int
-            Repeat the hyperparameter optimization by the specified number of
-            times and return the best result.
-        theta_jitter: float
-            Standard deviation of the random noise added to the initial
-            logscale hyperparameters across repeated optimization runs.
-        verbose: bool
-            Whether or not to print out the optimization progress and outcome.
-
-        Returns
-        -------
-        self: GaussianProcessRegressor
-            returns an instance of self.
-        ymean: 1D array, only if return_mean is True
-            Mean of the leave-one-out predictive distribution at query points.
-        std: 1D array, only if return_std is True
-            Standard deviation of the leave-one-out predictive distribution.
-        """
-        self.X = X
-        self.y = y
-
-        '''hyperparameter optimization'''
-        if self.optimizer:
-            opt = self._hyper_opt(
-                lambda theta, self=self: self.squared_loocv_error(
-                    theta, eval_gradient=True, clone_kernel=False,
-                    verbose=verbose
-                ),
-                self.kernel.theta.copy(),
-                tol, repeat, theta_jitter, verbose
-            )
-            if verbose:
-                print(f'Optimization result:\n{opt}')
-
-            if opt.success:
-                self.kernel.theta = opt.x
-            else:
-                raise RuntimeError(
-                    f'Minimum LOOCV optimization did not converge, got:\n'
-                    f'{opt}'
-                )
-
-        '''build and store GPR model'''
-        self.K = self._gramian(X)
-        self.Kinv = CholSolver(self.K)
-        self.Ky = self.Kinv @ self.y
-        if return_mean is False and return_std is False:
-            return self
-        else:
-            retvals = []
-            Kinv_diag = (self.Kinv @ np.eye(len(self.X))).diagonal()
-            if return_mean is True:
-                ymean = self.y - self.Kinv @ self.y / Kinv_diag
-                retvals.append(ymean * self.y_std + self.y_mean)
-            if return_std is True:
-                ystd = np.sqrt(1 / np.maximum(Kinv_diag, 1e-14))
-                retvals.append(ystd * self.y_std)
-            return (self, *retvals)
+    def fit_loocv(self, X, y, **options):
+        """Alias of :py:`fit_loocv(X, y, loss='loocv', **options)`."""
+        return self.fit(X, y, loss='loocv', **options)
 
     def predict(self, Z, return_std=False, return_cov=False):
         """Predict using the trained GPR model.
@@ -311,7 +272,7 @@ class GaussianProcessRegressor:
             z = (z - z_mean) / z_std
         else:
             z_mean, z_std = 0, 1
-        Kinv = CholSolver(self._gramian(Z))
+        Kinv, _ = self._invert(self._gramian(Z))
         Kinv_diag = (Kinv @ np.eye(len(Z))).diagonal()
         ymean = (z - Kinv @ z / Kinv_diag) * z_std + z_mean
         if return_std is True:
@@ -368,43 +329,40 @@ class GaussianProcessRegressor:
             kernel.theta = theta
 
         t_kernel = time.perf_counter()
+
         if eval_gradient is True:
             K, dK = self._gramian(X, kernel=kernel, jac=True)
         else:
             K = self._gramian(X, kernel=kernel)
+
         t_kernel = time.perf_counter() - t_kernel
 
         t_linalg = time.perf_counter()
-        try:
-            Kinv = CholSolver(K)
-        except np.linalg.LinAlgError as e:
-            raise np.linalg.LinAlgError(
-                'Kernel matrix is singular, try a larger `alpha` value.\n'
-                f'NumPy error information: {e}'
-            )
+
+        Kinv, logdet = self._invert(K)
         Ky = Kinv @ y
         yKy = y @ Ky
-        logdet = np.prod(np.linalg.slogdet(K))
-        t_linalg = time.perf_counter() - t_linalg
 
         if eval_gradient is True:
-            D_theta = np.zeros_like(theta)
-            for i, t in enumerate(theta):
-                dk = dK[:, :, i]
-                D_theta[i] = ((Kinv @ dk).trace() - Ky @ dk @ Ky) * np.exp(t)
-            retval = (yKy + logdet, D_theta)
+            d_theta = (
+                np.einsum('ij,ijk->k', Kinv @ np.eye(len(K)), dK) -
+                np.einsum('i,ijk,j', Ky, dK, Ky)
+            )
+            retval = (yKy + logdet, d_theta * np.exp(theta))
         else:
             retval = yKy + logdet
 
+        t_linalg = time.perf_counter() - t_linalg
+
         if verbose:
             mprint.table(
-                ('log(P)', '%12.5g', yKy + logdet),
-                ('yKy', '%12.5g', yKy),
-                ('logdet(K)', '%12.5g', logdet),
-                ('Norm(dK)', '%12.5g', np.linalg.norm(D_theta)),
+                ('logP', '%12.5g', yKy + logdet),
+                ('dlogP', '%12.5g', np.linalg.norm(d_theta)),
+                ('y^T.K.y', '%12.5g', yKy),
+                ('log|K| ', '%12.5g', logdet),
                 ('Cond(K)', '%12.5g', np.linalg.cond(K)),
-                ('t_GPU (s)', '%10.2g', t_kernel),
-                ('t_CPU (s)', '%10.2g', t_linalg),
+                ('GPU time', '%10.2g', t_kernel),
+                ('CPU time', '%10.2g', t_linalg),
             )
 
         return retval
@@ -464,12 +422,14 @@ class GaussianProcessRegressor:
         t_kernel = time.perf_counter() - t_kernel
 
         t_linalg = time.perf_counter()
-        Kinv = CholSolver(K) @ np.eye(len(X))
+
+        Kinv, logdet = self._invert(K)
+        if not isinstance(Kinv, np.ndarray):
+            Kinv = Kinv @ np.eye(len(X))
         Kinv_diag = Kinv.diagonal()
         Ky = Kinv @ y
         e = Ky / Kinv_diag
         squared_error = 0.5 * np.sum(e**2)
-        t_linalg = time.perf_counter() - t_linalg
 
         if eval_gradient is True:
             D_theta = np.zeros_like(theta)
@@ -484,11 +444,13 @@ class GaussianProcessRegressor:
         else:
             retval = squared_error
 
+        t_linalg = time.perf_counter() - t_linalg
+
         if verbose:
             mprint.table(
                 ('Sq.Err.', '%12.5g', squared_error),
-                ('logdet(K)', '%12.5g', np.prod(np.linalg.slogdet(K))),
-                ('Norm(dK)', '%12.5g', np.linalg.norm(D_theta)),
+                ('d(SqErr)', '%12.5g', squared_error),
+                ('log|K| ', '%12.5g', logdet),
                 ('Cond(K)', '%12.5g', np.linalg.cond(K)),
                 ('t_GPU (s)', '%10.2g', t_kernel),
                 ('t_CPU (s)', '%10.2g', t_linalg),
@@ -497,22 +459,16 @@ class GaussianProcessRegressor:
         return retval
 
     def _hyper_opt(self, fun, x0, tol, repeat, theta_jitter, verbose):
-        X0 = np.copy(self.kernel.theta)
         opt = None
 
         for r in range(repeat):
-            if r > 0:
-                x0 = X0 + theta_jitter * np.random.randn(len(X0))
-            else:
-                x0 = X0
-
             if verbose:
                 mprint.table_start()
 
             opt_local = minimize(
                 fun=fun,
                 method=self.optimizer,
-                x0=x0,
+                x0=x0 + theta_jitter * np.random.randn(len(x0)) if r else x0,
                 bounds=self.kernel.bounds,
                 jac=True,
                 tol=tol,
@@ -522,3 +478,44 @@ class GaussianProcessRegressor:
                 opt = opt_local
 
         return opt
+
+    def save(self, path, filename='model.pkl', overwrite=False):
+        """Save the trained GaussianProcessRegressor with the associated data
+        as a pickle.
+
+        Parameters
+        ----------
+        path: str
+            The directory to store the saved model.
+        filename: str
+            The file name for the saved model.
+        overwrite: bool
+            If True, a pre-existing file will be overwritten. Otherwise, a
+            runtime error will be raised.
+        """
+        f_model = os.path.join(path, filename)
+        if os.path.isfile(f_model) and not overwrite:
+            raise RuntimeError(
+                f'Path {f_model} already exists. To overwrite, set '
+                '`overwrite=True`.'
+            )
+        store = self.__dict__.copy()
+        store['theta'] = self.kernel.theta
+        store.pop('kernel', None)
+        pickle.dump(store, open(f_model, 'wb'), protocol=4)
+
+    def load(self, path, filename='model.pkl'):
+        """Load a stored GaussianProcessRegressor model from a pickle file.
+
+        Parameters
+        ----------
+        path: str
+            The directory where the model is saved.
+        filename: str
+            The file name for the saved model.
+        """
+        f_model = os.path.join(path, filename)
+        store = pickle.load(open(f_model, 'rb'))
+        theta = store.pop('theta')
+        self.__dict__.update(**store)
+        self.kernel.theta = theta
